"""
Categorization
==============

An example to illustrate binary categorizaiton with FreeDiscovery
"""

from __future__ import print_function

import os.path
import requests
import pandas as pd

pd.options.display.float_format = '{:,.3f}'.format
pd.options.display.expand_frame_repr = False

dataset_name = "treclegal09_2k_subset"     # see list of available datasets

BASE_URL = "http://localhost:5001/api/v0"  # FreeDiscovery server URL

#############################################################################
#
# 0. Load the example dataset
# ---------------------------

url = BASE_URL + '/example-dataset/{}'.format(dataset_name)
print(" GET", url)
input_ds = requests.get(url).json()


data_dir = input_ds['metadata']['data_dir']
dataset_definition = [{'document_id': row['document_id'],
                       'file_path': os.path.join(data_dir, row['file_path'])}
                      for row in input_ds['dataset']]

############################################################################
#
# 1. Feature extraction
# ---------------------

print("\n1.a Load dataset and initalize feature extraction")
url = BASE_URL + '/feature-extraction'
print(" POST", url)
res = requests.post(url).json()

dsid = res['id']
print("   => received {}".format(list(res.keys())))
print("   => dsid = {}".format(dsid))

###########################################################################
#
# 1.b Start feature extraction (in the background)

url = BASE_URL+'/feature-extraction/{}'.format(dsid)
print(" POST", url)
res = requests.post(url, json={'dataset_definition': dataset_definition})

#############################################################################
#
# 2. Calculate Latent Semantic Indexing
# -------------------------------------
# (used by Nearest Neighbors method)

url = BASE_URL + '/lsi/'
print("POST", url)

n_components = 100
res = requests.post(url,
                    json={'n_components': n_components,
                          'parent_id': dsid
                          }).json()

lsi_id = res['id']
print('  => LSI model id = {}'.format(lsi_id))
print(('  => SVD decomposition with {} dimensions explaining '
       '{:.2f} % variabilty of the data')
      .format(n_components, res['explained_variance']*100))

#############################################################################
#
# 3. Document categorization
# --------------------------
# 3.a. Train the categorization model

print("   {} positive, {} negative files".format(
      pd.DataFrame(input_ds['training_set'])
        .groupby('category').count()['document_id'], 0))

for method, use_lsi in [('LinearSVC', False),
                        ('NearestNeighbor', True)]:

    print('='*80, '\n', ' '*10,
          method, " + LSI" if use_lsi else ' ', '\n', '='*80)
    if use_lsi:
        # Categorization with the previously created LSI model
        parent_id = lsi_id
    else:
        # Categorization with original text features
        parent_id = dsid

    url = BASE_URL + '/categorization/'
    print(" POST", url)
    print(' Training...')

    res = requests.post(url,
                        json={'parent_id': parent_id,
                              'data': input_ds['training_set'],
                              'method': method,
                              'training_scores': True
                              }).json()

    mid = res['id']
    print("     => model id = {}".format(mid))
    print(("    => Training scores: MAP = {average_precision:.3f}, "
           "ROC-AUC = {roc_auc:.3f}, recall @20%: {recall_at_20p:.3f} ")
          .format(**res['training_scores']))

    print("\n3.b. Check the parameters used in the categorization model")
    url = BASE_URL + '/categorization/{}'.format(mid)
    print(" GET", url)
    res = requests.get(url).json()

    print('\n'.join(['     - {}: {}'.format(key, val)
          for key, val in res.items() if key not in ['index', 'category']]))

    print("\n3.c Categorize the complete dataset with this model")
    url = BASE_URL + '/categorization/{}/predict'.format(mid)
    print(" GET", url)
    res = requests.get(url, json={'subset': 'test'}).json()

    data = []
    for row in res['data']:
        nrow = {'document_id': row['document_id'],
                'category': row['scores'][0]['category'],
                'score': row['scores'][0]['score']}
        if method == 'NearestNeighbor':
            nrow['nearest_document_id'] = row['scores'][0]['document_id']
        data.append(nrow)

    df = pd.DataFrame(data).set_index('document_id')
    print(df)

    print("\n3.d Compute the categorization scores")
    url = BASE_URL + '/metrics/categorization'
    print(" GET", url)
    res = requests.post(url, json={'y_true': input_ds['dataset'],
                                   'y_pred': res['data']}).json()

    print(("    => Test scores: MAP = {average_precision:.3f}, "
           "ROC-AUC = {roc_auc:.3f}, recall @20%: {recall_at_20p:.3f} ")
          .format(**res))

#############################################################################
#
# 5 Delete the extracted features (and LSI decomposition)
url = BASE_URL + '/feature-extraction/{}'.format(dsid)
requests.delete(url)



# -*- coding: utf-8 -*-

import os
import pytest
import itertools
from unittest import SkipTest
import numpy as np
from numpy.testing import assert_equal, assert_array_equal
import pandas as pd

from ...utils import dict2type
from ...sklearn_compat import sklearn_version
from ...exceptions import OptionalDependencyMissing
from .base import (parse_res, V01, app, app_notest, get_features, get_features_lsi,
                   get_features_lsi_cached, get_features_cached)


#=============================================================================#
#
#                     LSI
#
#=============================================================================#

def test_api_lsi(app):
    dsid, pars, _ = get_features_cached(app)
    method = V01 + "/feature-extraction/{}".format(dsid)
    res = app.get(method)
    assert res.status_code == 200
    data = parse_res(res)
    method = V01 + "/lsi/"
    res = app.get_check(method, data=dict(parent_id=dsid,))

    lsi_pars = dict(n_components=101, parent_id=dsid)
    method = V01 + "/lsi/"
    data = app.post_check(method, json=lsi_pars)

    assert sorted(data.keys()) == ['explained_variance', 'id']
    lid = data['id']

    # checking again that we can load all the lsi models
    method = V01 + "/lsi/"
    data = app.get_check(method, data=dict(parent_id=dsid,))
    method = V01 + "/lsi/{}".format(lid)
    data = app.get_check(method)
    for key, vals in lsi_pars.items():
        assert vals == data[key]

    assert sorted(data.keys()) == sorted(["n_components", "parent_id"])

    for key in data.keys():
        assert data[key] == lsi_pars[key]


def _api_categorization_wrapper(app, solver, cv, n_categories,
                                n_categories_train=None, max_results=None,
                                subset='all', metric='cosine'):
    cv = (cv == 'cv')

    if 'CIRCLECI' in os.environ and cv == 1 and solver in ['LinearSVC',
                                                           'xgboost']:
        raise SkipTest  # Circle CI is too slow and timesout

    if solver.startswith('Nearest'):
        dsid, lsi_id, _, ds_input = get_features_lsi_cached(app, n_categories=n_categories)
        parent_id = lsi_id
    else:
        dsid, _, ds_input = get_features_cached(app, n_categories=n_categories)
        lsi_id = None
        parent_id = dsid

    method = V01 + "/feature-extraction/{}".format(dsid)
    data = app.get_check(method)

    categories_list = list(set([row['category'] for row in ds_input['dataset']]))

    if n_categories_train is None:
        training_set = ds_input['training_set']
    else:
        assert n_categories_train <= n_categories
        training_set = list(filter(lambda x: x['category'] in categories_list[:n_categories_train],
                            ds_input['training_set']))

    pars = {
          'parent_id': parent_id,
          'data': training_set,
          'method': solver,
          'cv': cv,
          'training_scores': True}

    method = V01 + "/categorization/"
    try:
        data = app.post_check(method, json=pars)
    except OptionalDependencyMissing:
        raise SkipTest

    assert dict2type(data) == {'id': 'str',
                               'training_scores': {'recall': 'float',
                                                   'f1': 'float',
                                                   'precision': 'float',
                                                   'roc_auc': 'float',
                                                   'average_precision': 'float',
                                                   'recall_at_20p': 'float'}}

    training_scores = data['training_scores']

    # it is very likely that there is an issue in the training scores
    if n_categories_train == 1:
        if sklearn_version < (0, 18, 0):
            pass
        else:
            # this yields wrong results for sklearn 0.17
            assert training_scores['f1'] > 0.99

    elif n_categories == 2:
        assert training_scores['average_precision'] > 0.73
        if solver == 'NearestNeighbor':
            assert training_scores['roc_auc'] > 0.7
        else:
            assert training_scores['roc_auc'] >= 0.5
    elif n_categories == 3 and solver == 'NearestNeighbor':
        assert training_scores['f1'] > 0.6
    else:
        assert training_scores['f1'] > 0.3

    mid = data['id']

    method = V01 + "/categorization/{}".format(mid)
    data = app.get_check(method)

    for key in ["method"]:
        assert pars[key] == data[key]

    method = V01 + "/categorization/{}/predict".format(mid)
    if max_results:
        json_data = {'max_results': max_results}
    else:
        json_data = {}
    json_data['subset'] = subset
    json_data['metric'] = metric

    data = app.get_check(method, json=json_data)
    data = data['data']
    response_ref = {'document_id': 'int',
                    'scores': [{'category': 'str',
                                'score': 'float',
                                }
                               ]}

    if solver == 'NearestNeighbor':
        response_ref['scores'][0]['document_id'] = 'int'

    if max_results is None:
        if subset == 'all':
            assert len(data) == len(ds_input['dataset'])
        elif subset == 'train':
            assert len(data) == len(training_set)
        elif subset == 'test':
            assert len(data) == len(ds_input['dataset']) - len(training_set)
        else:
            raise ValueError

    for row in data:
        assert dict2type(row) == response_ref

    if solver == 'NearestNeighbor':
        training_document_id = np.array([row['document_id'] for row in training_set])
        training_document_id_res = np.array([row['scores'][0]['document_id'] for row in data])
        assert_equal(np.in1d(training_document_id_res, training_document_id), True)

    method = V01 + "/metrics/categorization"
    data = app.post_check(method,
            json={'y_true': ds_input['dataset'],
                  'y_pred': data})

    assert dict2type(data) == {'precision': 'float',
                               'recall': 'float',
                               'f1': 'float',
                               'roc_auc': 'float',
                               'average_precision': 'float',
                               'recall_at_20p': 'float'}
    if n_categories == 2:
        assert data['average_precision'] > 0.7
        assert data['roc_auc'] > 0.7
        assert data['recall_at_20p'] > 0.2 # that's a very loose criterion
    else:
        if sklearn_version <= (0, 18, 0) and n_categories_train == 1: 
            pass
        else:
            # this yields wrong results for sklearn 0.17
            assert data['f1'] > 0.32

    method = V01 + "/categorization/{}".format(mid)
    res = app.delete_check(method)


_categoriazation_pars = itertools.product(["LinearSVC", "LogisticRegression",
                                           "NearestNeighbor", 'xgboost'],
                                          ['', 'cv'])

_categoriazation_pars = filter(lambda args: not (args[0].startswith('Nearest') and args[1] == 'cv'),
                               _categoriazation_pars)


@pytest.mark.parametrize("solver, cv", _categoriazation_pars)
def test_api_categorization_2cat(app, solver, cv):
    _api_categorization_wrapper(app, solver, cv, 2)


@pytest.mark.parametrize("n_categories", [1, 2, 3])
def test_api_categorization_2cat_unsupervised(app, n_categories):
    _api_categorization_wrapper(app, 'NearestNeighbor', cv='',
                                n_categories=n_categories,
                                n_categories_train=1)


@pytest.mark.parametrize("solver", ["LogisticRegression", "NearestNeighbor"])
def test_api_categorization_3cat(app, solver):
    _api_categorization_wrapper(app, solver, '', 3)


def test_api_categorization_max_results(app):
    _api_categorization_wrapper(app, 'LogisticRegression', '',
                                2, max_results=100)


@pytest.mark.parametrize("subset", ['all', 'test', 'train'])
def test_api_categorization_3cat(app, subset):
    _api_categorization_wrapper(app, 'LogisticRegression', '', 2, subset=subset)


def test_api_categorization_cosine_positive(app):
    _api_categorization_wrapper(app, 'NearestNeighbor', '', 2, metric='cosine-positive')


def test_api_categorization_subset_document_id(app):
    n_categories = 1
    dsid, lsi_id, _, ds_input = get_features_lsi_cached(app, n_categories=n_categories)
    method = V01 + "/feature-extraction/{}".format(dsid)
    data = app.get_check(method)

    training_set = ds_input['training_set']

    pars = {
          'parent_id': lsi_id,
          'data': training_set,
          'method': 'NearestNeighbor'}

    method = V01 + "/categorization/"
    data = app.post_check(method, json=pars)
    mid = data['id']

    method = V01 + "/categorization/{}/predict".format(mid)

    data = app.get_check(method, json={'batch_id': -1,
                                       'subset_document_id': [222784, 5184,
                                                              929296, 999999999]})
    data = data['data']
    assert len(data) == 3
    # check that only 3 subset documents are returned (9999999 is not a valid id)
    # and that they are in a correct order
    assert_array_equal([row['document_id'] for row in data], [5184, 222784, 929296])

    method = V01 + "/categorization/{}".format(mid)
    app.delete_check(method)


@pytest.mark.parametrize('sort_by', ['comp.graphics', 'rec.sport.baseball'])
def test_api_categorization_sort(app, sort_by):
    n_categories = 2
    dsid, lsi_id, _, ds_input = get_features_lsi_cached(app, n_categories=n_categories)
    method = V01 + "/feature-extraction/{}".format(dsid)
    data = app.get_check(method)

    training_set = ds_input['training_set']

    pars = {
          'parent_id': lsi_id,
          'data': training_set,
          'method': 'NearestNeighbor'}

    method = V01 + "/categorization/"
    data = app.post_check(method, json=pars)
    mid = data['id']

    method = V01 + "/categorization/{}/predict".format(mid)

    data = app.get_check(method, json={'batch_id': -1, "sort_by": sort_by})

    res = []
    for row in data['data']:
        res_el = {'document_id': row['document_id']}
        for scores in row['scores']:
            res_el[scores['category']] = scores['score']
        res.append(res_el)

    df = pd.DataFrame(res)
    df = df.set_index('document_id')

    if sort_by in df.columns:
        mask = pd.notnull(df[sort_by])
        assert_array_equal(df[mask].index.values,
                           df[mask].sort_values(sort_by, ascending=False).index.values)



# -*- coding: utf-8 -*-

from pathlib import Path
import os.path
from unittest import SkipTest
import re

import numpy as np
from numpy.testing import assert_allclose

import pandas as pd
import pytest
import itertools

from freediscovery.engine.pipeline import PipelineFinder
from freediscovery.engine.vectorizer import FeatureVectorizer
from freediscovery.engine.categorization import _CategorizerWrapper
from freediscovery.engine.lsi import _LSIWrapper
from freediscovery.io import parse_ground_truth_file
from freediscovery.metrics import categorization_score
from freediscovery.exceptions import OptionalDependencyMissing, WrongParameter
from freediscovery.tests.run_suite import check_cache


basename = Path(__file__).parent

cache_dir = check_cache()

EPSILON = 1e-4


data_dir = basename / ".." / ".." / "data" / "ds_001" / "raw"

ground_truth = parse_ground_truth_file(
                        str(data_dir / ".." / "ground_truth_file.txt"))

fe = FeatureVectorizer(cache_dir=cache_dir, mode='w')
vect_uuid = fe.setup()
fe.ingest(str(data_dir), file_pattern='.*\d.txt')


lsi = _LSIWrapper(cache_dir=cache_dir, parent_id=vect_uuid, mode='w')
lsi.fit_transform(n_components=6)


_test_cases = itertools.product(
                       [False, True],
                       ["LinearSVC", "LogisticRegression", 'xgboost',
                        "NearestNeighbor", "NearestCentroid"],
                       [None, 'fast'])

# 'MLPClassifier', 'ensemble-stacking' not supported in production the moment
_test_cases = filter(lambda x: not (x[1].startswith("Nearest") and x[2]),
                     _test_cases)


@pytest.mark.parametrize('use_lsi, method, cv', _test_cases)
def test_categorization(use_lsi, method, cv):

    if 'CIRCLECI' in os.environ and cv == 'fast'\
            and method in ['LinearSVC', 'xgboost']:
        raise SkipTest  # Circle CI is too slow and timesout

    if method == 'xgboost':
        try:
            import xgboost
        except ImportError:
            raise SkipTest

    if not use_lsi:
        uuid = vect_uuid
    else:
        uuid = lsi.mid

    cat = _CategorizerWrapper(cache_dir=cache_dir,
                              parent_id=uuid, cv_n_folds=2)
    cat.fe.db_.filenames_ = cat.fe.filenames_
    index = cat.fe.db_._search_filenames(ground_truth.file_path.values)

    try:
        model, Y_train = cat.fit(
                                index,
                                ground_truth.is_relevant.values,
                                method=method,
                                cv=cv)
    except OptionalDependencyMissing:
        raise SkipTest
    except WrongParameter:
        if method in ['NearestNeighbor', 'NearestCentroid']:
            return
        else:
            raise

    Y_pred, md = cat.predict()
    X_pred = np.arange(cat.fe.n_samples_, dtype='int')
    idx_gt = cat.fe.db_._search_filenames(ground_truth.file_path.values)

    scores = categorization_score(idx_gt,
                                  ground_truth.is_relevant.values,
                                  X_pred, np.argmax(Y_pred, axis=1))

    assert cat.get_params() is not None

    assert Y_pred.shape == (cat.fe.n_samples_,
                            len(np.unique(ground_truth.is_relevant.values)))

    if method == 'NearestNeighbor':
        assert md.shape == Y_pred.shape
    else:
        assert md is None

    if method in ['xgboost', 'ensemble-stacking']:
        # this parameter fail for some reason so far...
        return
    assert_allclose(scores['precision'], 1, rtol=0.5)
    assert_allclose(scores['recall'], 1, rtol=0.68)
    cat.delete()


def test_explain_categorization():
    from freediscovery.categorization import binary_sensitivity_analysis

    uuid = vect_uuid

    cat = _CategorizerWrapper(cache_dir=cache_dir, parent_id=uuid,
                              cv_n_folds=2)
    cat.fe.db_.filenames_ = cat.fe.filenames_
    index = cat.fe.db_._search_filenames(ground_truth.file_path.values)

    model, _ = cat.fit(index, ground_truth.is_relevant.values,
                       method='LogisticRegression')
    X = cat.fe._load_features()
    vect = cat.fe.vect_

    weights = binary_sensitivity_analysis(model, vect.vocabulary_, X[0, :])
    # not all vocabulary keys are returned
    assert len(weights.keys()) < len(vect.vocabulary_)


@pytest.mark.parametrize('n_steps', [2, 3])
def test_pipeline(n_steps):
    """ Test a 2 or 3 step pipelines with
        vectorizer (+ lsi) + classifier """

    if n_steps == 2:
        uuid = vect_uuid
    elif n_steps == 3:
        uuid = lsi.mid
    else:
        raise ValueError

    cat = _CategorizerWrapper(cache_dir=cache_dir,
                              parent_id=uuid, cv_n_folds=2)
    cat.fe.db_.filenames_ = cat.fe.filenames_
    index = cat.fe.db_._search_filenames(ground_truth.file_path.values)

    coefs, Y_train = cat.fit(index, ground_truth.is_relevant.values)

    cat.predict()

    assert len(cat.pipeline) == n_steps - 1

    # additional tests
    if n_steps == 3:
        pf = PipelineFinder.by_id(cat.mid, cache_dir)

        assert list(pf.keys()) == ['vectorizer', 'lsi', 'categorizer']
        assert list(pf.parent.keys()) == ['vectorizer', 'lsi']
        assert list(pf.parent.parent.keys()) == ['vectorizer']

        assert pf.mid == cat.mid
        assert pf.parent.mid == lsi.mid
        assert pf.parent.parent.mid == vect_uuid
        with pytest.raises(ValueError):
            pf.parent.parent.parent

        for estimator_type, mid in pf.items():
            path = str(pf.get_path(mid, absolute=False))
            if estimator_type == 'vectorizer':
                assert re.match('ediscovery_cache.*', path)
            elif estimator_type == 'lsi':
                assert re.match('ediscovery_cache.*lsi', path)
            elif estimator_type == 'categorizer':
                assert re.match('ediscovery_cache.*lsi.*categorizer', path)
            else:
                raise ValueError



# -*- coding: utf-8 -*-

# Author: Olivier Grisel <olivier.grisel@ensta.org>
# Copyright: 2012, Olivier Grisel
# License: BSD 3 clause
# (adapted from joblib 0.11)

import os
import tempfile

from sklearn.externals.joblib.pool import SYSTEM_SHARED_MEM_FS

def _get_temp_dir(pool_folder_name, temp_folder=None):
    """Get the full path to a subfolder inside the temporary folder.

    This function is originally used in joblib.pool.MemmapingPool, but
    it could also be used in combination with joblib.Memory, provided
    that the produced cache folders are manually cleaned to avoid
    running out of memory.

    Parameters
    ----------
    pool_folder_name : str
        Sub-folder name used for the serialization of a pool instance.

    temp_folder: str, optional
        Folder to be used by the pool for memmaping large arrays
        for sharing memory with worker processes. If None, this will try in
        order:

        - a folder pointed by the JOBLIB_TEMP_FOLDER environment
          variable,
        - /dev/shm if the folder exists and is writable: this is a
          RAMdisk filesystem available by default on modern Linux
          distributions,
        - the default system temporary folder that can be
          overridden with TMP, TMPDIR or TEMP environment
          variables, typically /tmp under Unix operating systems.

    Returns
    -------
    pool_folder : str
       full path to the temporary folder
    use_shared_mem : bool
       whether the temporary folder is written to tmpfs
    """
    use_shared_mem = False
    if temp_folder is None:
        temp_folder = os.environ.get('JOBLIB_TEMP_FOLDER', None)
    if temp_folder is None:
        if os.path.exists(SYSTEM_SHARED_MEM_FS):
            try:
                temp_folder = SYSTEM_SHARED_MEM_FS
                pool_folder = os.path.join(temp_folder, pool_folder_name)
                if not os.path.exists(pool_folder):
                    os.makedirs(pool_folder)
                use_shared_mem = True
            except IOError:
                # Missing rights in the the /dev/shm partition,
                # fallback to regular temp folder.
                temp_folder = None
    if temp_folder is None:
        # Fallback to the default tmp folder, typically /tmp
        temp_folder = tempfile.gettempdir()
    temp_folder = os.path.abspath(os.path.expanduser(temp_folder))
    pool_folder = os.path.join(temp_folder, pool_folder_name)
    return pool_folder, use_shared_mem



# Authors: Roman Yurchak
#
# License: BSD 3 clause

def emails_ignore_header(doc):
    """ Skip irrelevant email header fields """
    doc = '\n'.join([line for line in doc.splitlines()
                     if not line.strip().startswith(('To:', 'From:', 'Date:',
                                                     'Cc:', 'Sent:'))])
    return doc


processing_filters = {'emails_ignore_header': emails_ignore_header}



# Authors: Roman Yurchak
#
# License: BSD 3 clause

from textwrap import dedent


def jupyter_hide_code():
    """ Allow to toggle display on notebooks """
    import IPython.core.display as di

    # This line will hide code by default when the notebook is exported as HTML
    di.display_html(dedent("""<script>
                           jQuery(function() {
                           if (jQuery("body.notebook_app").length == 0) {
                                jQuery(".input_area").toggle();
                                jQuery(".prompt").toggle();
                            }});
                            </script>"""), raw=True)

    # This line will add a button to toggle visibility of code blocks,
    # for use with the HTML export version
    di.display_html('''<button onclick="jQuery('.input_area').toggle(); jQuery('.prompt').toggle();">Toggle code</button>''', raw=True)



# -*- coding: utf-8 -*-

import numpy as np
from numpy.testing import assert_allclose

from freediscovery.lsi import (_TruncatedSVD_LSI,
                               _compute_lsi_dimensionality)


def test_compute_lsi_dimensionality():
    n_components = 200
    n_components_res = _compute_lsi_dimensionality(n_components, 10000, 100000)
    assert n_components_res == n_components
    n_components_res = _compute_lsi_dimensionality(n_components, 500, 100000)
    assert n_components_res < n_components


def test_lsi_helper_class():
    import scipy.sparse

    X = scipy.sparse.rand(100, 10000)
    lsi = _TruncatedSVD_LSI(n_components=20)
    lsi.fit(X)
    X_p = lsi.transform_lsi(X)
    X_p2 = lsi.transform_lsi_norm(X)
    assert lsi.components_.shape == (20, X.shape[1])
    assert X_p.shape == (100, 20)
    assert X_p2.shape == (100, 20)


def test_lsi_book_example():
    """ LSI example taken from the "Information retrieval" (2004)
    book by Grossman & Ophir

    This illustrates the general principle of LSI using
    sklearn API with _TruncatedSVD_LSI
    """

    # replacing "a" with "aa" as the former seems
    # to be ignored by the CountVectorizer
    documents = ["Shipment of gold damaged in aa fire.",
                 "Delivery of silver arrived in aa silver truck.",
                 "Shipment of gold arrived in aa truck.",
                 ]
    querry = "gold silver truck"
    from sklearn.feature_extraction.text import CountVectorizer
    from sklearn.metrics.pairwise import cosine_similarity
    import scipy.linalg
    dm_vec = CountVectorizer()
    dm_vec.fit(documents)
    X = dm_vec.transform(documents)

    assert X.shape[1] == 11
    # checking the total number of elements in the document matrix
    assert X.sum() == 22

    q = dm_vec.transform([querry])

    lsi = _TruncatedSVD_LSI(n_components=2)  # algorithm='arpack')

    lsi.fit(X)
    X_p = lsi.transform_lsi(X)
    q_p = lsi.transform_lsi(q)

    U, s, Vh = scipy.linalg.svd(X.todense().T, full_matrices=False)

    q_p_2 = q.dot(U[:, :-1]).dot(np.diag(1./s[:-1]))
    assert_allclose(np.abs(q_p_2), np.array([[0.2140, 0.1821]]), 1e-3)
    X_p_2 = X.dot(U[:, :-1]).dot(np.diag(1./s[:-1]))

    assert_allclose(np.abs(np.abs(X_p_2)), np.abs(X_p))
    assert_allclose(np.abs(np.abs(q_p_2)), np.abs(q_p))

    D = cosine_similarity(X_p, q_p)

    assert_allclose(D[:2],
                    np.array([-0.05, 0.9910, 0.9543])[:2, None], 2e-2, 1e-2)



# -*- coding: utf-8 -*-
import sys

from freediscovery._version import __version__
from setuptools import setup, find_packages


with open('README.rst', 'rb') as fh:
    LONG_DESCRIPTION = fh.read().decode('utf-8')

with open('requirements.txt', 'rt') as fh:
    REQUIREMENTS = fh.read().splitlines()

with open('requirements_engine.txt', 'rt') as fh:
    REQUIREMENTS_ENGINE = fh.read().splitlines()


setup(name='freediscovery',
      version=__version__,
      description='Open source software for E-Discovery '
                  'and Information Retrieval',
      author='Grossman Labs',
      url="https://github.com/FreeDiscovery/FreeDiscovery",
      license='BSD',
      packages=find_packages(),
      include_package_data=True,
      long_description=LONG_DESCRIPTION,
      entry_points={
          'console_scripts': [
              'freediscovery = freediscovery.__main__:main'
          ]
      },
      classifiers=[
          'Development Status :: 4 - Beta',

          'Intended Audience :: Information Technology',
          'Intended Audience :: Legal Industry',

          'License :: OSI Approved :: BSD License',

          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',

          'Operating System :: OS Independent',

          'Topic :: Scientific/Engineering :: Artificial Intelligence',
          'Topic :: Text Processing :: General'],
      extras_require={'engine': REQUIREMENTS_ENGINE,
                      'core': REQUIREMENTS},
      keywords='information-retrieval machine-learning text-classification')



"""
Categorization Example [REST API]
---------------------------------

An example to illustrate binary categorizaiton with FreeDiscovery
"""

from __future__ import print_function

from time import time, sleep
from pathlib import Path
import os.path
from multiprocessing import Process
import requests
import pandas as pd
import sys

pd.options.display.float_format = '{:,.3f}'.format
pd.options.display.expand_frame_repr = False

dataset_name = "treclegal09_37k_subset"     # see list of available datasets

BASE_URL = "http://localhost:5001/api/v0"  # FreeDiscovery server URL

ingestion_method = 'file_path'


print(" 0. Load the example dataset")
url = BASE_URL + '/example-dataset/{}'.format(dataset_name)
print(" GET", url)
input_ds = requests.get(url).json()

data_dir = input_ds['metadata']['data_dir']
dataset_definition = []
for row in input_ds['dataset']:
    dd_row = {'document_id': row['document_id']}
    if ingestion_method == 'file_path':
        dd_row['file_path'] = os.path.join(data_dir, row['file_path'])
        dataset_definition.append(dd_row)
    elif ingestion_method == 'content':
        with Path(data_dir, row['file_path']).open('rt') as fh:
            try:
                dd_row['content'] = fh.read()
                dataset_definition.append(dd_row)
            except UnicodeDecodeError:
                print(Path(data_dir, row['file_path']), ' failed!')
    else:
        raise ValueError
# create a custom dataset definition for ingestion

# 1. Feature extraction

print("\n1.a Load dataset and initalize feature extraction")
url = BASE_URL + '/feature-extraction'
print(" POST", url)
res = requests.post(url, json={'max_df': 0.6, 'preprocess': ['emails_ignore_header']
                              }).json()

dsid = res['id']
print("   => received {}".format(list(res.keys())))
print("   => dsid = {}".format(dsid))

print("\n1.b Start feature extraction (in the background)")

# Make this call in a background process (there should be a better way of doing it)
url = BASE_URL+'/feature-extraction/{}'.format(dsid)
print(" POST", url)
t0 = time()
for i in range(0, len(dataset_definition), 8000):
    res = requests.post(url, json={'dataset_definition': dataset_definition[i:i + 8000],
                                   'vectorize': False})

print('     ... files ingested in {:.1f} s'.format((time() - t0)))
t1 = time()
res = requests.post(url, json={'vectorize': True})
print('     ... files vectorized in {:.1f} s'.format((time() - t0)))

print("\n1.d. check the parameters of the extracted features")
url = BASE_URL + '/feature-extraction/{}'.format(dsid)
print(' GET', url)
res = requests.get(url).json()

print('\n'.join(['     - {}: {}'.format(key, val)
      for key, val in res.items() if "filenames" not in key]))


# 4. Cleaning
print("\n5.a Delete the extracted features (and LSI decomposition)")
url = BASE_URL + '/feature-extraction/{}'.format(dsid)
print(" DELETE", url)
requests.delete(url)



"""
Categorization Interpretation Example
-------------------------------------

A visual interpretation for the binary categorization outcome for a single document
by looking at the relative contribution of individual words
"""
from __future__ import print_function

import os
from sklearn.datasets import fetch_20newsgroups
from sklearn.linear_model import LogisticRegression
from sklearn.feature_extraction.text import TfidfVectorizer
import matplotlib as mpl
mpl.use('Agg')

import matplotlib.pyplot as plt

from freediscovery.categorization import binary_sensitivity_analysis
from freediscovery.interpretation import explain_categorization, _make_cmap


newsgroups = fetch_20newsgroups(subset='train', categories=['sci.space', 'comp.graphics'],
                                remove=('headers', 'footers', 'quotes'))

document_id = 312  # the document id we want to visualize

vectorizer = TfidfVectorizer(stop_words='english')
X = vectorizer.fit_transform(newsgroups.data)

clf = LogisticRegression()
clf.fit(X, newsgroups.target)

repr_proba = 'Predicted: {0}: {{0:.2f}}, {1}: {{1:.2f}}'.format(*newsgroups.target_names)
print(repr_proba.format(*clf.predict_proba(X[document_id])[0]))
print('Actual label :', newsgroups.target_names[newsgroups.target[document_id]])


weights = binary_sensitivity_analysis(clf, vectorizer.vocabulary_, X[document_id])

cmap = _make_cmap(alpha=0.2, filter_ratio=0.15)
html, norm = explain_categorization(weights, newsgroups.data[document_id], cmap)

fig, ax = plt.subplots(1, 1, figsize=(6, 1.2))
plt.subplots_adjust(bottom=0.4, top=0.7)
cb1 = mpl.colorbar.ColorbarBase(ax, cmap=cmap, norm=norm, orientation='horizontal')

cb1.set_label('{} < ----- > {}'.format(*newsgroups.target_names))
ax.set_title('Relative word weights', fontsize=12)

# visualize the html results in sphinx gallery
tmp_dir = os.path.join('..', '..', 'doc', 'python', 'examples')
print(os.path.abspath(tmp_dir))
if os.path.exists(tmp_dir):
    with open(os.path.join(tmp_dir, 'out.html'), 'wt') as fh:
        fh.write(html)

####################################
# .. raw:: html
#     :file: out.html



# -*- coding: utf-8 -*-

import numpy as np
from numpy.testing import assert_allclose
import pytest
from sklearn.preprocessing import normalize

from freediscovery.metrics import (ratio_duplicates_score,
                                   f1_same_duplicates_score,
                                   recall_at_k_score,
                                   mean_duplicates_count_score,
                                   seuclidean_dist2cosine_sim)

def test_duplicate_metrics():
    x1 = np.array([0, 1, 1, 2, 3, 2])
    x2 = np.array([2, 3, 3, 4, 5, 4])  # just an ids renaming
    x3 = np.array([1, 2, 3, 4, 5, 6])  # no duplicates
    x4 = np.array([0, 1, 3, 2, 5, 2])  # half of original duplicates

    assert ratio_duplicates_score(x1, x2) == 1.0
    assert ratio_duplicates_score(x1, x3) == 0.0
    assert ratio_duplicates_score(x1, x4) == 0.5

    assert f1_same_duplicates_score(x1, x2) == 1.0
    assert f1_same_duplicates_score(x1, x3) == 0.0
    assert 0.25 < f1_same_duplicates_score(x1, x4) < 0.75  # more loose condition

    assert mean_duplicates_count_score(x1, x2) == 1.0
    assert mean_duplicates_count_score(x1, x3) == 0.5
    assert mean_duplicates_count_score(x1, x4) == 0.75

def test_euclidean2cosine():
    from sklearn.metrics.pairwise import pairwise_distances
    x = normalize([[0, 2, 3, 5]])
    y = normalize([[1, 3, 6, 7]])

    D_cos = pairwise_distances(x, y, metric='cosine')[0, 0]
    S_cos = 1 - D_cos
    D_seuc = pairwise_distances(x, y, metric='euclidean', squared=True)[0, 0]

    assert_allclose(S_cos, seuclidean_dist2cosine_sim(D_seuc))

def test_cosine2jaccard():
    from sklearn.metrics.pairwise import pairwise_distances
    from freediscovery.metrics import (cosine2jaccard_similarity,
                                       jaccard2cosine_similarity)

    x = np.array([[0, 0, 1., 1.]])
    y = np.array([[0, 1., 1., 0]])

    S_cos = 1 - pairwise_distances(x, y, metric='cosine')
    S_jac = cosine2jaccard_similarity(S_cos)
    S_jac_ref = 1 - pairwise_distances(x.astype('bool'), y.astype('bool'), metric='jaccard')

    assert_allclose(S_jac, S_jac_ref)

    S_cos2 = jaccard2cosine_similarity(S_jac)
    assert_allclose(S_cos2, S_cos)


@pytest.mark.parametrize('metric', ['cosine', 'jaccard', 'cosine_norm', 'jaccard_norm']) 
def test_cosine_jaccard_norm(metric):
    from freediscovery.metrics import _scale_cosine_similarity
    S_cos = 0.70710678

    S_res = _scale_cosine_similarity(S_cos, metric=metric)
    if metric == 'cosine':
        assert_allclose(S_res, S_cos)
    elif metric == 'jaccard':
        assert_allclose(S_res, 0.5, rtol=0.1)
    S_res2 = _scale_cosine_similarity(S_res, metric=metric, inverse=True)
    assert_allclose(S_res2, S_cos, rtol=0.1)


def test_cosine_positive():
    from freediscovery.metrics import _scale_cosine_similarity

    S_res = _scale_cosine_similarity(0.5, metric='cosine-positive')
    assert S_res == 0.5
    S_res = _scale_cosine_similarity(-0.5, metric='cosine-positive')
    assert S_res == 0


def test_recall_at_k_score():
    y_true = np.array([0, 1, 0, 1, 0])
    y_pred = np.array([0.2, 0.8, 0.1, 0.5, 0.6])  # just an ids renaming

    score = recall_at_k_score(y_true, y_pred, k=0)
    assert score == 0.0
    score = recall_at_k_score(y_true, y_pred, k=len(y_true))
    assert score == 1.0
    score = recall_at_k_score(y_true, y_pred, k=1)
    assert score == 0.5
    score = recall_at_k_score(y_true, y_pred, k=2)
    assert score == 0.5
    score = recall_at_k_score(y_true, y_pred, k=3)
    assert score == 1.0

    # test float input
    score = recall_at_k_score(y_true, y_pred, k=0.)
    assert score == 0.0
    score = recall_at_k_score(y_true, y_pred, k=0.2)
    assert score == 0.5
    score = recall_at_k_score(y_true, y_pred, k=1.0)
    assert score == 1.0



# -*- coding: utf-8 -*-

from unittest import SkipTest
import pytest

# adapted from https://github.com/seomoz/simhash-py/blob/master/test/test.py
jabberwocky = '''
    Twas brillig, and the slithy toves
      Did gyre and gimble in the wabe:
    All mimsy were the borogoves,
      And the mome raths outgrabe.
    "Beware the Jabberwock, my son!
      The jaws that bite, the claws that catch!
    Beware the Jubjub bird, and shun
      The frumious Bandersnatch!"
    He took his vorpal sword in hand:
      Long time the manxome foe he sought --
    So rested he by the Tumtum tree,
      And stood awhile in thought.
    And, as in uffish thought he stood,
      The Jabberwock, with eyes of flame,
    Came whiffling through the tulgey wood,
      And burbled as it came!
    One, two! One, two! And through and through
      The vorpal blade went snicker-snack!
    He left it dead, and with its head
      He went galumphing back.
    "And, has thou slain the Jabberwock?
      Come to my arms, my beamish boy!
    O frabjous day! Callooh! Callay!'
      He chortled in his joy.
    `Twas brillig, and the slithy toves
      Did gyre and gimble in the wabe;
    All mimsy were the borogoves,
      And the mome raths outgrabe.'''
jabberwocky_author = ' - Lewis Carroll (Alice in Wonderland)'


def test_simhash():

    try:
        from simhash import num_differing_bits
    except ImportError:
        raise SkipTest
    from sklearn.feature_extraction.text import HashingVectorizer
    from freediscovery.near_duplicates import SimhashNearDuplicates

    DISTANCE = 4

    fe = HashingVectorizer(ngram_range=(4, 4), analyzer='word')

    X = fe.fit_transform([jabberwocky,
                          jabberwocky + jabberwocky_author,
                          jabberwocky_author,
                          jabberwocky])

    sh = SimhashNearDuplicates()
    sh.fit(X)

    # make sure small changes in the text results in a small number of different bytes
    assert num_differing_bits(*sh._fit_shash[:2]) <= 3
    # different text produces a large number of different bytes
    assert num_differing_bits(*sh._fit_shash[1:3]) >= 20

    # same text produces a zero bit difference
    assert num_differing_bits(*sh._fit_shash[[0, -1]]) == 0

    simhash, cluster_id, dup_pairs = sh.query(distance=DISTANCE, blocks=42)
    assert str(dup_pairs.dtype) == 'uint64'
    assert str(cluster_id.dtype) == 'int64'
    assert str(dup_pairs.dtype) == 'uint64'

    assert simhash[0] == simhash[-1]        # duplicate documents have the same simhash
    assert cluster_id[0] == cluster_id[-1]  # and belong to the same cluster

    for idx, shash in enumerate(simhash):
        if (shash == simhash).sum() == 1:  # ignore duplicates
            assert sh.get_index_by_hash(shash) == idx

    for pairs in dup_pairs:
        assert num_differing_bits(*pairs) <= DISTANCE


@pytest.mark.parametrize('n_rand_lexicons,', [1, 5, 100])
def test_imatch(n_rand_lexicons):

    from sklearn.feature_extraction.text import TfidfVectorizer
    from freediscovery.near_duplicates import IMatchNearDuplicates

    DISTANCE = 4

    fe = TfidfVectorizer(ngram_range=(4, 4), analyzer='word',
                         min_df=0.25, max_df=0.75)

    X = fe.fit_transform([jabberwocky,
                          jabberwocky + jabberwocky_author,
                          jabberwocky_author,
                          jabberwocky])

    sh = IMatchNearDuplicates(n_rand_lexicons=n_rand_lexicons)
    sh.fit(X)

    assert sh.labels_.shape[0] == X.shape[0]
    assert sh.hash_.shape[0] == X.shape[0]
    assert sh.hash_is_dup_.shape[0] == X.shape[0]

    # different documents produce different hash
    assert sh.labels_[0] != sh.labels_[2]

    # same text produces same hash
    assert sh.labels_[0] == sh.labels_[-1]

    # RY: not sure what other tests could be run for I-Match



# Authors: Roman Yurchak
#
# License: BSD 3 clause


class RankerMixin(object):
    """Mixin class for all ranking estimators in FreeDiscovery.
    A ranker is a binary classifier without a decision threshold.
    """
    # so that thing would still work with scikit learn
    _estimator_type = "ranker"

    def score(self, X, y, sample_weight=None):
        """Returns the ROC score of the prediction.
        Best possible score is 1.0 and the worst in 0.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Test samples.
        y : array-like, shape = (n_samples)
            True values for X.
        sample_weight : array-like, shape = [n_samples], optional
            Sample weights.
        Returns
        -------
        score : float
            ROC score of self.decision_function(X) wrt. y.
        """

        from .metrics import roc_auc_score
        return roc_auc_score(y, self.decision_function(X),
                             sample_weight=sample_weight,)



from .ranking import auc
from .ranking import average_precision_score
from .ranking import precision_recall_curve
from .ranking import roc_auc_score
from .ranking import roc_curve
from .ranking import dcg_score
from .ranking import ndcg_score




# Authors: William Fagan
# Copyright (c) 2013-2017 William Fagan
# License: The MIT License (MIT)

import sys
from gzip import GzipFile
from io import BytesIO

from flask import request, current_app


if sys.version_info[:2] == (2, 6):
    class GzipFile(GzipFile):
        """ Backport of context manager support for python 2.6"""
        def __enter__(self):
            if self.fileobj is None:
                raise ValueError("I/O operation on closed GzipFile object")
            return self

        def __exit__(self, *args):
            self.close()


class DictCache(object):

    def __init__(self):
        self.data = {}

    def get(self, key):
        return self.data.get(key)

    def set(self, key, value):
        self.data[key] = value


class Compress(object):
    """
    The Compress object allows your application to use Flask-Compress.

    When initialising a Compress object you may optionally provide your
    :class:`flask.Flask` application object if it is ready. Otherwise,
    you may provide it later by using the :meth:`init_app` method.

    :param app: optional :class:`flask.Flask` application object
    :type app: :class:`flask.Flask` or None
    """
    def __init__(self, app=None):
        """
        An alternative way to pass your :class:`flask.Flask` application
        object to Flask-Compress. :meth:`init_app` also takes care of some
        default `settings`_.

        :param app: the :class:`flask.Flask` application object.
        """
        self.app = app
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        defaults = [
            ('COMPRESS_MIMETYPES', ['text/html', 'text/css', 'text/xml',
                                    'application/json',
                                    'application/javascript']),
            ('COMPRESS_LEVEL', 6),
            ('COMPRESS_MIN_SIZE', 500),
            ('COMPRESS_CACHE_KEY', None),
            ('COMPRESS_CACHE_BACKEND', None),
            ('COMPRESS_REGISTER', True),
        ]

        for k, v in defaults:
            app.config.setdefault(k, v)

        backend = app.config['COMPRESS_CACHE_BACKEND']
        self.cache = backend() if backend else None
        self.cache_key = app.config['COMPRESS_CACHE_KEY']

        if (app.config['COMPRESS_REGISTER'] and
                app.config['COMPRESS_MIMETYPES']):
            app.after_request(self.after_request)

    def after_request(self, response):
        app = self.app or current_app
        accept_encoding = request.headers.get('Accept-Encoding', '')

        if (response.mimetype not in app.config['COMPRESS_MIMETYPES'] or
            'gzip' not in accept_encoding.lower() or
            not 200 <= response.status_code < 300 or
            (response.content_length is not None and
             response.content_length < app.config['COMPRESS_MIN_SIZE']) or
            'Content-Encoding' in response.headers):
            return response

        response.direct_passthrough = False

        if self.cache:
            key = self.cache_key(response)
            gzip_content = self.cache.get(key) or self.compress(app, response)
            self.cache.set(key, gzip_content)
        else:
            gzip_content = self.compress(app, response)

        response.set_data(gzip_content)

        response.headers['Content-Encoding'] = 'gzip'
        response.headers['Content-Length'] = response.content_length

        vary = response.headers.get('Vary')
        if vary:
            if 'accept-encoding' not in vary.lower():
                response.headers['Vary'] = '{}, Accept-Encoding'.format(vary)
        else:
            response.headers['Vary'] = 'Accept-Encoding'

        return response

    def compress(self, app, response):
        gzip_buffer = BytesIO()
        with GzipFile(mode='wb',
                      compresslevel=app.config['COMPRESS_LEVEL'],
                      fileobj=gzip_buffer) as gzip_file:
            gzip_file.write(response.get_data())
        return gzip_buffer.getvalue()



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import pytest
import json
import re
from unittest import SkipTest
import numpy as np
from numpy.testing import assert_equal, assert_almost_equal

from .. import fd_app
from ...utils import _silent, dict2type, sdict_keys

from .base import (parse_res, V01, app, app_notest, get_features_cached,
                   get_features_lsi_cached)


def _check_htree_consistency(htree, dataset_size):
    all_cluster_id = [row['cluster_id'] for row in htree]
    for row in htree:
        assert np.in1d(row['children'], all_cluster_id).all()
    row1 = htree[0]

    max_tree_depth = np.max([row['cluster_depth'] for row in htree])
    for level in range(max_tree_depth):
        out = []
        for srow in filter(lambda row: row['cluster_depth'] == level, htree):
            out += [k['document_id'] for k in srow['documents']]

        if False:
            # this is disabled by default with prune_single_clusters=False
            assert len(out) == dataset_size, "depth={}".format(level)
        # no duplicate ids at the same hierarchy level
        assert len(np.unique(out)) == len(out)
    if len(row1['documents']) > 1:
        assert_almost_equal(np.mean([doc['similarity']
                            for doc in row1['documents']]),
                            row1['cluster_similarity'])
        assert np.max([doc['similarity'] for doc in row1['documents']]) > \
            row1['cluster_similarity']



# =============================================================================#
#
#                     Clustering
#
# =============================================================================#


@pytest.mark.parametrize("model, use_lsi, n_clusters, optimal_sampling",
                         [('k-mean', False, 13, False),
                          ('birch', True, -1, False),
                          ('birch', True, -1, True),
                          ('birch', True, 13, False),
                          ('dbscan', True, None, False)])
def test_api_clustering(app, model, use_lsi, n_clusters, optimal_sampling):

    dsid, lsi_id, _, _ = get_features_lsi_cached(app)
    if use_lsi:
        parent_id = lsi_id
    else:
        lsi_id = None
        parent_id = dsid

    pars = {'parent_id': parent_id, }

    if model == 'birch' and n_clusters <= 0:
        is_hierarchical = True
    else:
        is_hierarchical = False

    method = V01 + "/feature-extraction/{}".format(dsid)
    data = app.get_check(method)

    url = V01 + "/clustering/" + model
    if model != 'dbscan':
        pars['n_clusters'] = n_clusters
    if model == 'dbscan':
        pars.update({'eps': 0.1, "min_samples": 2})
    elif model == 'birch':
        pars['min_similarity'] = 0.7
    data = app.post_check(url, json=pars)

    assert sorted(data.keys()) == sorted(['id'])
    mid = data['id']

    url += '/{}'.format(mid)
    n_top_words = 2
    pars = {'n_top_words': n_top_words}
    if optimal_sampling:
        pars['return_optimal_sampling'] = True
    data = app.get_check(url, query_string=pars)

    assert dict2type(data, max_depth=1) == {'data': 'list'}
    for row in data['data']:
        ref_res = {'cluster_id': 'int', 'cluster_similarity': 'float',
                   'documents': 'list', 'cluster_size': 'int'}
        if is_hierarchical and not optimal_sampling:
            ref_res['children'] = 'list'
            ref_res['cluster_depth'] = 'int'
        if not optimal_sampling:
            ref_res['cluster_label'] = 'str'
            assert re.match('[^\[]+', row['cluster_label'])
        assert dict2type(row, max_depth=1) == ref_res
        # make sure we have space separated words, not a str(list)
        for irow in row['documents']:
            assert dict2type(irow) == {'document_id': 'int',
                                       'similarity': 'float'}
    if model != 'dbscan' and not is_hierarchical:
        assert len(data['data']) == 13

    app.delete_check(url)

def _get_min_similarity(data):
    min_sim = 1.0
    for row in data:
        c_min_sim = min(sdoc['similarity']
                        for sdoc in row['documents'])
        min_sim = min(c_min_sim, min_sim)
    return min_sim


def test_clustering_birch_cosine_positive(app):
    dsid, lsi_id, _, _ = get_features_lsi_cached(app, hashed=False)
    parent_id = lsi_id
    url = V01 + "/clustering/birch"
    pars = {'min_similarity': 0.5, 'parent_id': parent_id, 'metric': 'cosine'}
    mid = app.post_check(url, json=pars)['id']
    data = app.get_check(url + '/' + mid)
    assert _get_min_similarity(data['data']) < 0.0
    pars['metric'] = 'cosine-positive'
    mid = app.post_check(url, json=pars)['id']
    data = app.get_check(url + '/' + mid)
    assert _get_min_similarity(data['data']) == 0.0
    pars['metric'] = 'cosine-p'
    with pytest.raises(ValueError):
        mid = app.post_check(url, json=pars)['id']


def test_clustering_max_tree_depth(app):
    dsid, lsi_id, a, b = get_features_lsi_cached(app, hashed=False)
    parent_id = lsi_id
    url = V01 + "/clustering/birch"
    pars = {'min_similarity': 0.8, 'parent_id': parent_id,
            'branching_factor': 10}
    data = app.post_check(url, json=pars)
    mid = data['id']
    data = app.get_check(url + '/' + mid)
    max_depth = max(el['cluster_depth'] for el in data['data'])
    max_depth_lim = 2
    assert max_depth > max_depth_lim

    pars['max_tree_depth'] = max_depth_lim

    mid = app.post_check(url, json=pars)['id']
    data2 = app.get_check(url + '/' + mid)

    assert max(el['cluster_depth'] for el in data2['data']) == max_depth_lim

    dataset_size = len(b['dataset'])

    _check_htree_consistency(data['data'], dataset_size)
    _check_htree_consistency(data2['data'], dataset_size)



# -*- coding: utf-8 -*-

import warnings

from sklearn.externals import joblib

from freediscovery.engine.base import _BaseWrapper
from freediscovery.search import Search


class _SearchWrapper(_BaseWrapper):
    """ Document search wrapper

    Parameters
    ----------
    cache_dir : str
      folder where the model will be saved
    parent_id : str, optional
      dataset id
    mid : str, optional
      model id
    """

    _wrapper_type = "search"

    def __init__(self, cache_dir='/tmp/',  parent_id=None, mid=None):

        super(_SearchWrapper, self).__init__(cache_dir=cache_dir,
                                             parent_id=parent_id,
                                             mid=mid)

    def search(self, text, internal_id=None, metric='cosine'):
        """
        Search given some text query

        Parameters
        ----------
        text : str
          the query string. This will be ignored if an
          internal_id parameter is provided.
        internal_id : int
          the document id used as a search query.
          If provided the text input will be ignored.
        metric : str
          the output metric to use
        """
        if internal_id is not None:
            if 'lsi' not in self.pipeline:
                warnings.warn('Search using a document_id as a query'
                              ' should not be applied in the space of '
                              ' raw document term vectors due'
                              ' to the curse of dimensionality.'
                              ' Please add an LSI processing'
                              ' step (i.e. use `parent_id=lsi_id`)')
            vect = None
            tfidf = None
        else:
            vect = self.fe.vect_
            vect.set_params(input='content')
            tfidf = self.fe.tfidf_

        X = self.pipeline.data

        if "lsi" in self.pipeline:
            lsi = joblib.load(str(
                              self.pipeline.get_path(self.pipeline['lsi']) /
                              'model'))
        else:
            lsi = None

        s = Search(vect, tfidf, lsi)
        s.fit(X)

        if internal_id is not None:
            dist = s.search_id(internal_id, metric=metric)
        else:
            dist = s.search(text, metric=metric)

        return dist



# -*- coding  utf-8 -*-

import numpy as np
from sklearn.externals import joblib

from freediscovery.engine.base import _BaseWrapper
from freediscovery.utils import setup_model

from freediscovery.externals import jwzthreading as jwzt


class _EmailThreadingWrapper(_BaseWrapper):
    """ JWZ Email threading class


    Parameters
    ----------
    cache_dir : str
      folder where the model will be saved
    parent_id : str, optional
      dataset id
    mid : str, optional
      model id
    """

    _wrapper_type = "threading"

    def __init__(self, cache_dir='/tmp/',  parent_id=None, mid=None,
                 decode_header=False):

        super(_EmailThreadingWrapper, self).__init__(cache_dir=cache_dir,
                                                     parent_id=parent_id,
                                                     mid=mid,
                                                     load_model=True)

        if not (self.fe.dsid_dir / 'email_metadata').exists():
            raise ValueError('The email metadata was not found. Please rerun'
                             ' feature extraction with '
                             '`parse_email_headers=True` option')

    def thread(self, index=None, group_by_subject=False,
               sort_by_key='message_idx', sort_missing=-1):
        """
        Thread the emails

        Parameters
        ----------
        index : array-like, shape (n_samples)
           document indices of the training set

        Returns
        -------
        cmod : sklearn.BaseEstimator
           the scikit learn classifier object
        Y_train : array-like, shape (n_samples)
           training predictions
        group_by_subject : boolean, default=True
           group emails by subject

        Returns
        -------

        tree : array (N_samples)
           the id of the parent element in the tree
        root_idx : array (N_samples)
           the id of the root element in the tree
        """
        if index is None:
            index = np.arange(self.fe.n_samples_)

        d_all = joblib.load(str(self.fe.dsid_dir / 'email_metadata'))

        threads = jwzt.thread(d_all, group_by_subject)

        threads = [el.collapse_empty() for el in threads]

        threads = jwzt.sort_threads(threads, key=sort_by_key,
                                    missing=sort_missing)

        cmod = threads

        mid, mid_dir = setup_model(self.model_dir)

        pars = {
            'group_by_subject': group_by_subject
        }
        self._pars = pars
        joblib.dump(pars, str(mid_dir / 'pars'))
        joblib.dump(cmod, str(mid_dir / 'model'))

        self.mid = mid
        self.cmod = cmod
        return cmod



# -*- coding: utf-8 -*-

import os.path

import pytest

from freediscovery.engine.vectorizer import FeatureVectorizer
from freediscovery.engine.lsi import _LSIWrapper
from freediscovery.engine.search import _SearchWrapper

from freediscovery.tests.run_suite import check_cache

basename = os.path.dirname(__file__)
cache_dir = check_cache()
data_dir = os.path.join(basename, "..", "..", "data", "ds_001", "raw")


@pytest.mark.parametrize('kind,', ['regular', 'semantic'])
def test_search_wrapper(kind):
    # check for syntax errors etc in the wrapper

    fe = FeatureVectorizer(cache_dir=cache_dir, mode='w')
    vect_uuid = fe.setup()
    fe.ingest(data_dir, file_pattern='.*\d.txt')

    if kind == 'semantic':
        lsi = _LSIWrapper(cache_dir=cache_dir, parent_id=vect_uuid, mode='w')
        lsi.fit_transform(n_components=20)
        parent_id = lsi.mid
    else:
        parent_id = vect_uuid

    sw = _SearchWrapper(cache_dir=cache_dir, parent_id=parent_id)
    dist = sw.search("so that I can reserve a room")
    assert dist.shape == (fe.n_samples_,)
    # document 1 found by
    # grep -rn "so that I can reserve a room"
    # freediscovery/data/ds_001/raw/
    assert dist.argmax() == 1



# -*- coding: utf-8 -*-
import os
import os.path
import shutil
import pytest


from freediscovery.engine.stop_words import _StopWordsWrapper
from freediscovery.tests.run_suite import check_cache


from freediscovery.stop_words import CUSTOM_STOP_WORDS as csw
from freediscovery.stop_words import COMMON_FIRST_NAMES as cfns

cache_dir = check_cache()

tested_stop_words = ['one', 'two', 'three', 'four', 'five', 'six']


@pytest.mark.parametrize('csw_name, csw_list', [
                        ('test_acstw', tested_stop_words),
                        ('common_first_names', cfns),
                        ('csw', csw)
                         ])
def test_custom_stop_words_param(csw_name, csw_list):
    """Test to save, retrieve and delete the custom stop words

       tested_stop_words - list of custom stop words,
       cfns - list of common first names imported from stop_words.py
              to use as a custom stop words in testing
    """

    # test to save and retrieve the custom stop words
    custom_sw = _StopWordsWrapper(cache_dir=cache_dir)
    custom_sw.save(name=csw_name, stop_words=csw_list)
    custom_stop_words = custom_sw.load(name=csw_name)
    assert(custom_stop_words == csw_list)
    assert(len(custom_stop_words) == len(csw_list))
    i = 0
    for word in custom_stop_words:
        assert word == csw_list[i]
        i += 1

    # test to delete the custom stop words
    custom_sw.delete(name=csw_name)
    with pytest.raises(FileNotFoundError):
        custom_sw.load(name=csw_name)

    # clearing from temporary testing data
    if os.path.isdir(os.path.join(cache_dir, 'stop_words')):
        shutil.rmtree(os.path.join(cache_dir, 'stop_words'), ignore_errors=True)



# -*- coding: utf-8 -*-

import os.path
import warnings

import pytest
from freediscovery.utils import _split_path
from freediscovery.datasets import _normalize_cachedir
from .run_suite import check_cache


def test_split_path():
    assert _split_path('abc/test4') == ['abc', 'test4']
    assert _split_path('abc/test4/') == ['abc', 'test4']
    if os.name == 'nt':  # on windows
        assert _split_path('C:\\abc\\test4') == ["C:\\", 'abc', 'test4']
        assert _split_path('C:\\abc\\test4\\') == ["C:\\", 'abc', 'test4']
        assert _split_path('/test/test4') == ['\\', 'test', 'test4']
    else:
        assert _split_path('/test/test4') == ['/', 'test', 'test4']
        # this raises an error on windows for some reason
        assert _split_path('//abc/test4/') == ["//", 'abc', 'test4']


def test_normalize_cachedir():
    assert str(_normalize_cachedir('/tmp/')) == os.path.normpath('/tmp/ediscovery_cache')
    assert str(_normalize_cachedir('/tmp/ediscovery_cache')) == os.path.normpath('/tmp/ediscovery_cache')


@pytest.yield_fixture(autouse=True, scope='session')
def test_suite_cleanup():
    import shutil
    import traceback
    # setup
    yield
    # teardown - put your command here
    cache_dir = check_cache()
    try:
        shutil.rmtree(cache_dir)
        print('\nSucessfully removed cache_dir={}'.format(cache_dir))
    except Exeption as e:
        warnings.warn('Failed to remove cache_dir={}, \n{}'.format(
                      cache_dir, traceback.print_tb(e.__traceback__)))



# -*- coding: utf-8 -*-

CUSTOM_STOP_WORDS = ['com']

# List of common english first names
# source http://names.mongabay.com/male_names.htm
# source http://names.mongabay.com/female_names.htm
COMMON_FIRST_NAMES = [ "james","john","robert","michael","william","david",
"richard","charles","joseph","thomas","christopher","daniel","paul","mark",
"donald","george","kenneth","steven","edward","brian","ronald","anthony",
"kevin","jason","matthew","gary","timothy","jose","larry","jeffrey","frank",
"scott","eric","stephen","andrew","raymond","gregory","joshua","jerry","dennis",
"walter","patrick","peter","harold","douglas","henry","carl","arthur","ryan",
"roger","joe","juan","jack","albert","jonathan","justin","terry","gerald","keith",
"samuel","willie","ralph","lawrence","nicholas","roy","benjamin","bruce","brandon",
"adam","harry","fred","wayne","billy","steve","louis","jeremy","aaron","randy",
"howard","eugene","carlos","russell","bobby","victor","martin","ernest","phillip",
"todd","jesse","craig","alan","shawn","clarence","sean","philip","chris","johnny",
"earl","jimmy","antonio","danny","bryan","tony","luis","mike","stanley","leonard",
"nathan","dale","manuel","rodney","curtis","norman","allen","marvin","vincent",
"glenn","jeffery","travis","jeff","chad","jacob","lee","melvin","alfred","kyle",
"francis","bradley","jesus","herbert","frederick","ray","joel","edwin","don",
"eddie","ricky","troy","randall","barry","alexander","bernard","mario","leroy",
"francisco","marcus","micheal","theodore","clifford","miguel","oscar","jay","jim",
"tom","calvin","alex","jon","ronnie","bill","lloyd","tommy","leon","derek","warren",
"darrell","jerome","floyd","leo","alvin","tim","wesley","gordon","dean","greg",
"jorge","dustin","pedro","derrick","dan","lewis","zachary","corey","herman",
"maurice","vernon","roberto","clyde","glen","hector","shane","ricardo","sam",
"rick","lester","brent","ramon","charlie","tyler","gilbert","gene","marc",
"reginald","ruben","brett","angel","nathaniel","rafael","leslie","edgar","milton",
"raul","ben","chester","cecil","duane","franklin","andre","elmer","brad","gabriel",
"ron","mitchell","roland","arnold","harvey","jared","adrian","karl","cory","claude",
"erik","darryl","jamie","neil","jessie","christian","javier","fernando","clinton",
"ted","mathew","tyrone","darren","lonnie","lance","cody","julio","kelly","kurt",
"allan","nelson","guy","clayton","hugh","max","dwayne","dwight","armando","felix",
"jimmie","everett","jordan","ian","wallace","ken","bob","jaime","casey","alfredo",
"alberto","dave","ivan","johnnie","sidney","byron","julian","isaac","morris","clifton",
"willard","daryl","ross","virgil","andy","marshall","salvador","perry","kirk",
"sergio","marion","tracy","seth","kent","terrance","rene","eduardo","terrence",
"enrique","freddie","wade", "mary","patricia","linda","barbara","elizabeth",
"jennifer","maria","susan","margaret","dorothy","lisa","nancy","karen","betty",
"helen","sandra","donna","carol","ruth","sharon","michelle","laura","sarah",
"kimberly","deborah","jessica","shirley","cynthia","angela","melissa","brenda",
"amy","anna","rebecca","virginia","kathleen","pamela","martha","debra","amanda",
"stephanie","carolyn","christine","marie","janet","catherine","frances","ann",
"joyce","diane","alice","julie","heather","teresa","doris","gloria","evelyn",
"jean","cheryl","mildred","katherine","joan","ashley","judith","rose","janice",
"kelly","nicole","judy","christina","kathy","theresa","beverly","denise","tammy",
"irene","jane","lori","rachel","marilyn","andrea","kathryn","louise","sara",
"anne","jacqueline","wanda","bonnie","julia","ruby","lois","tina","phyllis",
"norma","paula","diana","annie","lillian","emily","robin","peggy","crystal",
"gladys","rita","dawn","connie","florence","tracy","edna","tiffany","carmen",
"rosa","cindy","grace","wendy","victoria","edith","kim","sherry","sylvia",
"josephine","thelma","shannon","sheila","ethel","ellen","elaine","marjorie",
"carrie","charlotte","monica","esther","pauline","emma","juanita","anita",
"rhonda","hazel","amber","eva","debbie","april","leslie","clara","lucille",
"jamie","joanne","eleanor","valerie","danielle","megan","alicia","suzanne",
"michele","gail","bertha","darlene","veronica","jill","erin","geraldine",
"lauren","cathy","joann","lorraine","lynn","sally","regina","erica","beatrice",
"dolores","bernice","audrey","yvonne","annette","june","samantha","marion",
"dana","stacy","ana","renee","ida","vivian","roberta","holly","brittany",
"melanie","loretta","yolanda","jeanette","laurie","katie","kristen","vanessa",
"alma","sue","elsie","beth","jeanne","vicki","carla","tara","rosemary","eileen",
"terri","gertrude","lucy","tonya","ella","stacey","wilma","gina","kristin",
"jessie","natalie","agnes","vera","willie","charlene","bessie","delores",
"melinda","pearl","arlene","maureen","colleen","allison","tamara","joy",
"georgia","constance","lillie","claudia","jackie","marcia","tanya","nellie",
"minnie","marlene","heidi","glenda","lydia","viola","courtney","marian",
"stella","caroline","dora","jo","vickie","mattie","terry","maxine","irma",
"mabel","marsha","myrtle","lena","christy","deanna","patsy","hilda","gwendolyn",
"jennie","nora","margie","nina","cassandra","leah","penny","kay","priscilla",
"naomi","carole","brandy","olga","billie","dianne","tracey","leona","jenny",
"felicia","sonia","miriam","velma","becky","bobbie","violet","kristina","toni",
"misty","mae","shelly","daisy","ramona","sherri","erika","katrina","claire",
"lindsey","lindsay","geneva","guadalupe","belinda","margarita","sheryl","cora",
"faye","ada","natasha","sabrina","isabel","marguerite","hattie","harriet",
"molly","cecilia","kristi","brandi","blanche","sandy","rosie","joanna","iris",
"eunice","angie","inez","lynda","madeline","amelia","alberta","genevieve",
"monique","jodi","janie","maggie","kayla","sonya","jan","lee","kristine",
"candace","fannie","maryann","opal","alison","yvette","melody","luz","susie",
"olivia","flora","shelley","kristy","mamie","lula","lola","verna","beulah",
"antoinette","candice","juana","jeannette","pam","kelli","hannah","whitney",
"bridget","karla","celia","latoya","patty","shelia","gayle","della","vicky",
"lynne","sheri","marianne","kara","jacquelyn","erma","blanca","myra","leticia",
"pat","krista","roxanne","angelica","johnnie","robyn","francis","adrienne",
"rosalie","alexandra","brooke","bethany","sadie","bernadette","traci","jody",
"kendra","jasmine","nichole","rachael","chelsea","mable","ernestine","muriel",
"marcella","elena","krystal","angelina","nadine","kari","estelle","dianna",
"paulette","lora","mona","doreen","rosemarie","angel","desiree","antonia",
"hope","ginger","janis","betsy","christie","freda","mercedes","meredith",
"lynette","teri","cristina","eula","leigh","meghan","sophia","eloise",
"rochelle","gretchen","cecelia","raquel","henrietta","alyssa","jana","kelley",
"gwen","kerry","jenna","tricia","laverne","olive","alexis","tasha","silvia",
"elvira","casey","delia","sophie","kate","patti","lorena","kellie","sonja",
"lila","lana","darla","may","mindy","essie","mandy","lorene","elsa","josefina",
"jeannie","miranda","dixie","lucia","marta","faith","lela","johanna","shari",
"camille","tami","shawna","elisa","ebony","melba","ora","nettie","tabitha",
"ollie","jaime","winifred","kristie","marina","alisha","aimee","rena","myrna",
"marla","tammie","latasha","bonita","patrice","ronda","sherrie","addie",
"francine","deloris","stacie","adriana","cheri","shelby","abigail","celeste",
"jewel","cara","adele","rebekah","lucinda","dorthy","chris","effie","trina",
"reba","shawn","sallie","aurora","lenora","etta","lottie","kerri","trisha",
"nikki","estella","francisca","josie","tracie","marissa","karin","brittney",
"janelle","lourdes","laurel","helene","fern","elva","corinne","kelsey","ina",
"bettie","elisabeth","aida","caitlin","ingrid","iva","eugenia","christa",
"goldie","cassie","maude","jenifer","therese","frankie","dena","lorna","janette",
"latonya","candy","morgan","consuelo","tamika","rosetta","debora","cherie",
"polly","dina","jewell","fay","jillian","dorothea","nell","trudy","esperanza",
"patrica","kimberley","shanna","helena","carolina","cleo","stefanie","rosario",
"ola","janine","mollie","lupe","alisa","lou","maribel","susanne","bette",
"susana","elise","cecile","isabelle","lesley","jocelyn","paige","joni",
"rachelle","leola","daphne","alta","ester","petra","graciela","imogene",
"jolene","keisha","lacey","glenna","gabriela","keri","ursula","lizzie",
"kirsten","shana","adeline","mayra","jayne","jaclyn","gracie","sondra",
"carmela","marisa","rosalind","charity","tonia","beatriz","marisol","clarice",
"jeanine","sheena","angeline","frieda","lily","robbie","shauna","millie",
"claudette","cathleen","angelia","gabrielle","autumn","katharine","summer",
"jodie","staci","lea","christi","jimmie","justine","elma","luella","margret",
"dominique","socorro","rene","martina","margo","mavis","callie","bobbi",
"maritza","lucile","leanne","jeannine","deana","aileen","lorie","ladonna",
"willa","manuela","gale","selma","dolly","sybil","abby","lara","dale","ivy",
"dee","winnie","marcy","luisa","jeri","magdalena","ofelia","meagan","audra",
"matilda","leila","cornelia","bianca","simone","bettye","randi","virgie",
"latisha","barbra","georgina","eliza","leann","bridgette","rhoda","haley",
"adela","nola","bernadine","flossie","ila","greta","ruthie","nelda","minerva",
"lilly","terrie","letha","hilary","estela","valarie","brianna","rosalyn",
"earline","catalina","ava","mia","clarissa","lidia","corrine","alexandria",
"concepcion","tia","sharron","rae","dona","ericka","jami","elnora","chandra",
"lenore","neva","marylou","melisa","tabatha","serena","avis","allie","sofia",
"jeanie","odessa","nannie","harriett","loraine","penelope","milagros","emilia",
"benita","allyson","ashlee","tania","tommie","esmeralda","karina","eve",
"pearlie","zelma","malinda","noreen","tameka","saundra","hillary","amie",
"althea","rosalinda","jordan","lilia","alana","gay","clare","alejandra",
"elinor","michael","lorrie","jerri","darcy","earnestine","carmella","taylor",
"noemi","marcie","liza","annabelle","louisa","earlene","mallory","carlene",
"nita","selena","tanisha","katy","julianne","john","lakisha","edwina",
"maricela","margery","kenya","dollie","roxie","roslyn","kathrine","nanette",
"charmaine","lavonne","ilene","kris","tammi","suzette","corine","kaye","jerry",
"merle","chrystal","lina","deanne","lilian","juliana","aline","luann","kasey",
"maryanne","evangeline","colette","melva","lawanda","yesenia","nadia","madge",
"kathie","eddie","ophelia","valeria","nona","mitzi","mari","georgette",
"claudine","fran","alissa","roseann","lakeisha","susanna","reva","deidre",
"chasity","sheree","carly","james","elvia","alyce","deirdre","gena","briana",
"araceli","katelyn","rosanne","wendi","tessa","berta","marva","imelda",
"marietta","marci","leonor","arline","sasha","madelyn","janna","juliette",
"deena","aurelia","josefa","augusta","liliana","young","christian","lessie",
"amalia","savannah","anastasia","vilma","natalia","rosella","lynnette",
"corina","alfreda","leanna","carey","amparo","coleen","tamra","aisha","wilda",
"karyn","cherry","queen","maura","mai","evangelina","rosanna","hallie","erna",
"enid","mariana","lacy","juliet","jacklyn","freida","madeleine","mara","hester",
"cathryn","lelia","casandra","bridgett","angelita","jannie","dionne","annmarie",
"katina","beryl","phoebe","millicent","katheryn","diann","carissa","maryellen",
"liz","lauri","helga","gilda","adrian","rhea","marquita","hollie","tisha",
"tamera","angelique","francesca","britney","kaitlin","lolita","florine",
"rowena","reyna","twila","fanny","janell","ines","concetta","bertie","alba",
"brigitte","alyson","vonda","pansy","elba","noelle","letitia","kitty","deann",
"brandie","louella","leta","felecia","sharlene","lesa","beverley","robert",
"isabella","herminia","terra","celina"
]



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
#from __future__ import unicode_literals

import sys
import os
import os.path
from unittest import SkipTest

import numpy as np
from numpy.testing import assert_allclose, assert_equal
import scipy.sparse
import itertools
import pytest
from textwrap import dedent


def test_count_duplicates():
    from freediscovery.utils import _count_duplicates

    x = np.array([1, 2, 1, 2, 2, 0])

    y = _count_duplicates(x)
    assert_equal(y, np.array([2, 3, 2, 3, 3, 1]))


def test_docstring_description():
    from freediscovery.utils import _docstring_description
    from freediscovery.datasets import load_dataset

    res = _docstring_description(dedent(load_dataset.__doc__))

    assert len(res.splitlines()) == 24

def test_dictkey2type():
    from freediscovery.utils import dict2type

    assert dict2type('djsk')  == 'str'
    assert dict2type(['t', 1]) == ['str', 'int']
    assert dict2type({'t': {'b': 0.1}}) == {'t': {'b': 'float'}}

    # make sure we don't change the original object
    x = {'x': {'a': 3}}
    dict2type(x)
    assert x == {'x': {'a': 3}}



# -*- coding: utf-8 -*-

import os
import pytest
import pandas as pd
import numpy as np
import itertools
from numpy.testing import (assert_array_equal, assert_array_less,
                           assert_allclose)
from sklearn.externals import joblib
from freediscovery.externals.sklearn.metrics import dcg_score

from ...utils import dict2type

from .base import (parse_res, V01, app, get_features_cached,
                   get_features_lsi_cached, CACHE_DIR)


@pytest.mark.parametrize("method, min_score, max_results",
                         itertools.product(['regular', 'semantic'],
                                           [-1, 0.1],
                                           [None, 1000000, 3, 0]))
def test_search(app, method, min_score, max_results):

    dsid, lsi_id, _, input_ds = get_features_lsi_cached(app)
    if method == 'semantic':
        parent_id = lsi_id
    elif method == 'regular':
        lsi_id = None
        parent_id = dsid
    query = """The American National Standards Institute sells ANSI standards, and also
    ISO (international) standards.  Their sales office is at 1-212-642-4900,
    mailing address is 1430 Broadway, NY NY 10018.  It helps if you have the
    complete name and number.
    """
    query_file_document_id = 2256

    pars = dict(parent_id=parent_id,
                min_score=min_score,
                query=query)
    if max_results is not None:
        pars['max_results'] = max_results

    data = app.post_check(V01 + "/search/", json=pars)
    assert sorted(data.keys()) == ['data', 'pagination']
    data = data['data']
    for row in data:
        assert dict2type(row) == {'score': 'float',
                                  'document_id': 'int'}
    df = pd.DataFrame(data)
    assert (np.diff(df.score.values) <= 0).all()
    assert_array_less(min_score, df.score.values)
    if max_results:
        assert len(df) <= min(max_results, len(input_ds['dataset']))
    elif min_score > -1:
        pass
    else:
        assert len(df) == 1967

    id_mapping = app.post_check(V01 + "/feature-extraction/{}/id-mapping"
                                .format(dsid))
    id_mapping = pd.DataFrame(id_mapping['data']).set_index('document_id')

    # check that relevant documents get returned first
    res = df.set_index('document_id').merge(id_mapping, left_index=True, right_index=True)
    res['document_id_new'] = res.file_path.str.extract('(\d+)', expand=True).astype('int')
    res['rank_position'] = np.arange(len(res))
    res = res.set_index('document_id_new')
    res['ground_truth'] = np.in1d(res.index.values, [1190, 2256]).astype('int')
    assert dcg_score(res.ground_truth.values, res.score.values) == 1.0
    assert res.loc[query_file_document_id].rank_position == 0

    # compare with pre-computed cosine similarity
    if method == 'regular':
        assert_allclose(res.loc[[2256, 2378, 1975], 'score'].values,
                        [0.5660, 0.1589, 0.1140],
                        rtol=1e-3)

def test_search_retrieve_batch(app):
    dsid, lsi_id, _, input_ds = get_features_lsi_cached(app)
    parent_id = lsi_id

    max_results = -1
    query_document_id = 3844
    total_document_number = 1967 - 1  # removing the query one

    for batch_id, batch_size in [(-1, 1000),
                                 (0, 983),
                                 (1, 983),
                                 (2, 983),
                                 (1, 1000)]:

        pars = dict(parent_id=parent_id,
                    max_results=max_results,
                    sort=True,
                    query_document_id=query_document_id,
                    batch_id=batch_id,
                    batch_size=batch_size)

        data = app.post_check(V01 + "/search/", json=pars)
        assert sorted(data.keys()) == ['data', 'pagination']
        for row in data['data']:
            assert dict2type(row) == {'score': 'float',
                                      'document_id': 'int'}
        assert dict2type(data['pagination']) == {
                                  'total_response_count': 'int',
                                  'current_response_count': 'int',
                                  'batch_id': 'int',
                                  'batch_id_last': 'int'}
        scores = np.array([row['score'] for row in data['data']])
        assert (np.diff(scores) <=  0).all()
        assert len(data['data']) == data['pagination']['current_response_count']
        assert data['pagination']['total_response_count'] == total_document_number
        assert data['pagination']['batch_id'] == batch_id
        if batch_id == -1:
            assert len(data['data']) == total_document_number
            assert data['pagination']['batch_id_last'] == batch_id
        elif batch_id >= 0:
            assert data['pagination']['current_response_count'] <= batch_size
            if batch_id == 1 and batch_size == 983:
                assert data['pagination']['current_response_count'] == batch_size
            elif batch_id == 2:
                assert data['pagination']['current_response_count'] == 0


def test_search_document_id(app):
    dsid, lsi_id, _, input_ds = get_features_lsi_cached(app, hashed=False)
    parent_id = lsi_id

    max_results = 2
    query_document_id = 3844

    pars = dict(parent_id=parent_id,
                max_results=max_results,
                sort=True,
                query_document_id=query_document_id)

    data = app.post_check(V01 + "/search/", json=pars)
    assert sorted(data.keys()) == ['data', 'pagination']
    data = data['data']
    for row in data:
        assert dict2type(row) == {'score': 'float',
                                  'document_id': 'int'}
    scores = np.array([row['score'] for row in data])
    assert (np.diff(scores) <= 0).all()
    assert len(data) == min(max_results, len(input_ds['dataset']))
    # assert data[0]['document_id'] == query_document_id
    # assert data[0]['score'] >= 0.99


def test_search_subset_document_id(app):
    dsid, lsi_id, _, input_ds = get_features_lsi_cached(app, hashed=False)
    parent_id = lsi_id

    query_document_id = 3844

    pars = dict(parent_id=parent_id,
                sort=True,
                query_document_id=query_document_id)

    # last document id is invalid
    pars['subset_document_id'] = [3705625, 1110916, 999999999]
    data = app.post_check(V01 + "/search/", json=pars)
    data = data['data']
    # check that the response only includes documents from the subset,
    # in the correct order
    assert_array_equal([row['document_id'] for row in data],
                       [1110916, 3705625])


def test_search_consistency(app):
    """ A number of consistency checks"""
    import pickle
    from scipy.stats import pearsonr
    from sklearn.metrics.pairwise import cosine_similarity
    dataset_id, lsi_id, ds_pars, input_ds = get_features_lsi_cached(app)
    query_document_id = 1

    input_ds = pd.DataFrame(input_ds['dataset']).set_index('document_id')

    # compute semantic search
    pars = dict(parent_id=lsi_id,
                query_document_id=query_document_id)
    data = app.post_check(V01 + "/search/", json=pars)
    df_s = pd.DataFrame(data['data']).set_index('document_id')
    df_s = df_s.merge(input_ds, how='left', left_index=True, right_index=True)

    # manually compute the similarity for a pair of documents and
    # check that it's the same as the one computed by the system
    with open(os.path.join(CACHE_DIR, 'ediscovery_cache', dataset_id, 'vectorizer'), 'rb') as fh:
        vect = pickle.load(fh)
    lsi_est = joblib.load(os.path.join(CACHE_DIR, 'ediscovery_cache', dataset_id,
                          'lsi', lsi_id, 'model'))
    X_tmp = []
    comp_document_id = 2365444
    for document_id in [query_document_id, comp_document_id]:
        with open(os.path.join(ds_pars['data_dir'],
            input_ds.loc[document_id].file_path), 'rb') as fh:
            X_tmp.append(fh.read().decode('utf-8'))
    X_tmp = vect.transform(X_tmp)
    X_tmp = lsi_est.transform_lsi_norm(X_tmp)
    assert_allclose(cosine_similarity(X_tmp[[0]], X_tmp[[1]]),
                    df_s.loc[comp_document_id].score)

    # check that providing the query_document_id or the same document as text
    # produces the same results
    with open(os.path.join(ds_pars['data_dir'],
                           input_ds.loc[query_document_id].file_path), 'rt') as fh:
        query_txt = fh.read()
    pars = dict(parent_id=lsi_id,
                query=query_txt)
    data = app.post_check(V01 + "/search/", json=pars)
    df_s_txt = pd.DataFrame(data['data']).set_index('document_id')
    df_s_txt = df_s_txt.merge(input_ds, how='left',
                              left_index=True, right_index=True)
    df_s_txt.loc[query_document_id].score == 1.0
    df_s_txt = df_s_txt[df_s_txt.index != query_document_id]
    assert_allclose(df_s.score.values, df_s_txt.score.values)

    # check that query document is on average closer to the documents of its class
    query_category = input_ds.loc[query_document_id].category
    assert df_s[df_s.category == query_category].score.quantile(q=0.75) \
        > df_s[df_s.category != query_category].score.quantile(q=0.75)


def test_search_eq_categorization(app):
    """Check that NN categorization with a single training document and semantic search
    returns the same results
    """
    dataset_id, lsi_id, ds_pars, input_ds = get_features_lsi_cached(app)
    query_document_id = 1

    input_ds = pd.DataFrame(input_ds['dataset']).set_index('document_id')

    # compute regular search
    pars = dict(parent_id=lsi_id,
                query_document_id=query_document_id)
    data = app.post_check(V01 + "/search/", json=pars)
    df_s = pd.DataFrame(data['data']).set_index('document_id')

    pars = {
          'parent_id': lsi_id,
          'data': [{'document_id': query_document_id, 'category': 'search'}],
          'method': 'NearestNeighbor',
          'subset': 'test'
           }

    method = V01 + "/categorization/"
    data = app.post_check(method, json=pars)

    method = V01 + "/categorization/{}/predict".format(data['id'])
    data = app.get_check(method)
    categories = [el['scores'][0]['category'] for el in data['data']]
    assert categories == ['search']*len(categories)

    df_c = pd.DataFrame([{'document_id': row['document_id'],
                          'score': row['scores'][0]['score']}
                         for row in data['data']]).set_index('document_id')

    # The order do not appear to be exactly matching (probably due to tie-breaking)
    assert (df_c.index == df_s.index).sum() / df_s.shape[0] > 0.994
    assert_allclose(df_c.score.values, df_s.score.values)


def test_ingestion_no_document_id_provided(app):
    """ Test what happens if no document id are provided during ingestion
    We use search merely because it's the simplest processing that can be done
    in a single API call"""
    data_dir = os.path.dirname(__file__)
    data_dir = os.path.join(data_dir, "..", "..", "data", "ds_001", "raw")
    method = V01 + "/feature-extraction/"
    data = app.post_check(method)
    dsid = data['id']
    app.post_check(method + dsid, json={'data_dir': data_dir})

    data = app.post_check(V01 + "/search/", json={'parent_id': dsid,
                                                  'query': 'test'})
    assert dict2type(data['data'][0]) == {'score': 'float',
                                          'document_id': 'int'}



# -*- coding: utf-8 -*-

from freediscovery.utils import dict2type
from .base import parse_res, V01, app, app_notest, get_features, get_features_lsi

# ============================================================================#
#
#                     Custom Stop Words
#
# ============================================================================#

def test_stop_words(app):
    name = "test_acstw"
    tested_stop_words = ['one', 'two', 'three', 'foure', 'five', 'six']

    method = V01 + "/stop-words/"
    pars = dict(name=name, stop_words=tested_stop_words)
    data = app.post_check(method, json=pars)

    method = V01 + "/stop-words/{}".format(name)
    data = app.get_check(method)

    assert dict2type(data, collapse_lists=True) == {'name': 'str', 'stop_words': ['str']}
    assert data["stop_words"] == tested_stop_words

    method = V01 + "/stop-words/{}".format(name)
    app.delete_check(method)



import sys
import multiprocessing

# Adapter from
# http://stackoverflow.com/questions/3041986/apt-command-line-interface-like-yes-no-input


def _query_yes_no(question, default="yes", overwrite=False):
    """Ask a yes/no question via input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    if overwrite:
        return overwrite
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")


class _TeeLogger(object):
    """ tee stdout & stderr to a file """
    def __init__(self, name, mode='at'):
        self.file = open(name, mode)
        self.stdout = sys.stdout
        self.stderr = sys.stderr
        sys.stdout = self
        sys.stderr = self

    def flush(self):
        self.file.flush()
        self.stdout.flush()
        self.stderr.flush()

    def __del__(self):
        sys.stdout = self.stdout
        sys.stderr = self.stderr
        self.file.close()

    def write(self, data):
        self.file.write(data)
        self.stdout.write(data)


def _number_of_workers(max_workers=6):
    """ Compute the number of worker processes for gunicorn"""
    num_w = (multiprocessing.cpu_count() * 2) + 1
    if num_w > max_workers:
        num_w = max_workers
    return num_w



# -*- coding: utf-8 -*-

import os
from collections import OrderedDict

from freediscovery.exceptions import ModelNotFound
from freediscovery.utils import _split_path
from pathlib import Path

from sklearn.externals import joblib


class PipelineFinder(OrderedDict):
    """Walk through the hierarchy of existing models
    and find the processing pipeline that terminate with the
    model having the given uid

    Parameters
    ----------
    uid : str
      a unique model id
    cache_dir : str
      folder where models are saved

    Returns
    -------
    result : OrderedDict
      the prior processing pipeline with as keys the processing step type
      and as values the model ids
    """

    def __init__(self, mid=None, cache_dir="/tmp/", steps=None):
        self._loaded_models = {}

        self.mid = mid
        if steps is None:
            steps = OrderedDict()

        self.cache_dir = self._normalize_cachedir(cache_dir)

        super(PipelineFinder, self).__init__(steps)

    @staticmethod
    def _normalize_cachedir(cache_dir):
        """ Normalize the cachedir path. This ensures that the cache_dir
        ends with "ediscovery_cache"
        """
        cache_dir = os.path.normpath(str(cache_dir))
        if 'ediscovery_cache' not in cache_dir:  # not very pretty
            cache_dir = os.path.join(cache_dir, "ediscovery_cache")
        return Path(cache_dir)

    @classmethod
    def by_id(cls, mid, cache_dir="/tmp/"):
        """ Find a pipeline by id

        Parameters
        ----------
        mid : str
          a unique model id
        cache_dir : str
          folder where models are saved

        Returns
        -------
        result : OrderedDict
          the prior processing pipeline with as keys the processing step type
          and as values the model ids
        """
        cache_dir = cls._normalize_cachedir(cache_dir)

        pipeline = cls(mid=mid, cache_dir=cache_dir)

        cache_dir_base = cache_dir.parent
        _break_flag = False
        for root, subdirs, files in os.walk(str(cache_dir)):
            root = os.path.relpath(root, str(cache_dir_base))
            for sdir in subdirs:
                path = os.path.join(root, sdir)
                path_hierarchy = _split_path(path)
                if len(path_hierarchy) % 2 == 1:
                    # the path is of the form
                    # ['ediscovery_cache']
                    # or ['ediscovery_cache', 'ce196de4c7de4e57', 'cluster']
                    # ignore it
                    continue

                if path_hierarchy[-1] == mid:
                    # found the model
                    _break_flag = True
                    break
            if _break_flag:
                break
        else:
            raise ModelNotFound('Model id {} not found in {}!'.format(mid, cache_dir))

        if path_hierarchy[0] == 'ediscovery_cache':
            path_hierarchy[0] = 'vectorizer'
        else:
            raise ValueError('path_hierarchy should start with ediscovery_cache',
                             'this indicates a bug in the code')

        for idx in range(len(path_hierarchy)//2):
            key, val = path_hierarchy[2*idx], path_hierarchy[2*idx+1]
            if key in pipeline:
                raise NotImplementedError('The current PipelineFinder class does not support'
                                          'multiple identical processing steps'
                                          'duplicates of {} found!'.format(key))
            pipeline[key] = val
        return pipeline

    @property
    def parent(self):
        """ Make a new pipeline without the latest node """

        if len(self.keys()) <= 1:
            raise ValueError("Can't take the parent of a root node!")

        # create a copy
        steps = OrderedDict(self)
        # get all the steps except the last one
        steps.popitem(last=True)

        return PipelineFinder(mid=list(steps.values())[-1],
                              cache_dir=self.cache_dir, steps=steps)

    @property
    def data(self):
        """ Load the data provided by the last node of the pipeline """
        last_node = list(self.keys())[-1]
        ds_path = self.get_path(self[last_node])

        if last_node == "vectorizer":
            full_path = ds_path / 'features'
        elif last_node == 'lsi':
            full_path = ds_path / 'data'
        return joblib.load(str(full_path))

    def get_path(self, mid=None, absolute=True):
        """ Find the path to the model specified by mid """
        import itertools

        if mid is None:
            mid = self.mid

        if mid not in self.values():
            raise ValueError('{} is not a processing step current pipeline,\n {}'.format(mid, self)) 
        idx = list(self.values()).index(mid)
        valid_keys = list(self.keys())[:idx+1]
        path = list(itertools.chain.from_iterable(
                            [[key, self[key]] for key in valid_keys]))
        if absolute:
            del path[0]  # "ediscovery_cache" is already present in cache_dir
            rel_path = os.path.join(*path)
            return self.cache_dir / rel_path
        else:
            path[0] = 'ediscovery_cache'
            return Path(os.path.join(*path))
